#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "mpi.h"

#define MSG_REPS 10
#define DEF_BUFFLEN 64
#define MTAG 99

#define BUFF_VAL(i) ((i)+1) /* expected value in ith element of buffer */

int main(int argc, char *argv[]) {
  int rank, nprocs;
  double *mbuf; int len, reps;
  int i, d, r;
  double t1, t2, tbar;
  int nErrs;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  len = (argc > 1)? atoi(argv[1]): DEF_BUFFLEN;
  reps = (len < 100? 100: (len < 1000? 10: 1)) * MSG_REPS;
  reps = (argc > 2)? atoi(argv[2]): reps;

  if ((nprocs & (nprocs-1)) != 0) {
    if (rank == 0)
      printf("one-all SF on hypercube: nprocs=%d must be a power of 2\n", 
	     nprocs);
    exit(1);
  }
  for (d=0; (1 << d) < nprocs; d++); // d = lg2(nprocs)  

  mbuf = calloc(len*nprocs, sizeof(double));
  if (rank == 0) // for checking, ensure different values on root
    for (i=0; i < len; i++)
      mbuf[i] = BUFF_VAL(i);

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Barrier(MPI_COMM_WORLD);
  t1 = MPI_Wtime();
  MPI_Barrier(MPI_COMM_WORLD);
  tbar = MPI_Wtime() - t1;

  if (rank==0)
    printf("one-all SF on hypercube dim %d: %d words %d reps; t_bar=%.2es\n", 
	   d, len, reps, tbar);

  t1 = 0.0;
  for (r=0; r < reps+1; r++) {
    if (r==1) t1 = MPI_Wtime(); //exclude rep 0 for warm-up
    //assert (nprocs == 2);
    if (rank == 0)
      MPI_Send(mbuf, len, MPI_DOUBLE, rank+1, MTAG, MPI_COMM_WORLD);
    if (rank == 1)
      MPI_Recv(mbuf, len, MPI_DOUBLE, rank-1, MTAG, MPI_COMM_WORLD,
	       MPI_STATUS_IGNORE);
  } //for(r...)

  MPI_Barrier(MPI_COMM_WORLD);
  t2 = MPI_Wtime() - t1 - tbar;
  if (rank==0)
    printf("time: av t=%.2es bw=%.2eMB/s\n", t2/reps, 1.0E-6*len*reps/t2);

  nErrs = 0;
  for (i=0; i < len; i++)
    nErrs += (mbuf[i] != BUFF_VAL(i));
  if (nErrs > 0)
    printf("process %d: %d incorrect elements in buffer!\n", rank, nErrs);

  free(mbuf);
  MPI_Finalize();
  return 0;
} //main()
